// main.c

// DSPIC33FJ16MC102 Configuration Bit Settings

// 'C' source line config statements

// CONFIG2
#pragma config POSCMOD = NONE           // Primary Oscillator Select (Primary oscillator disabled)
#pragma config ALTI2C = OFF             // Alternate I2C pins (I2C mapped to SDA1/SCL1)
#pragma config LPOL = ON                // Motor Control PWM Low Side Polarity bit (PWM module low side output pins have active-high output polarity)
#pragma config IOL1WAY = ON             // IOLOCK Protection (Allow Only One Re-configuration)
#pragma config OSCIOFNC = OFF           // Primary Oscillator Output Function (OSC2 pin has clock out function)
#pragma config FCKSM = CSDCMD           // Clock Switching and Monitor (Clock switching and Fail-Safe Clock Monitor are disabled)
#pragma config FNOSC = FRCPLL          // Oscillator Mode (Internal Fast RC (FRC) with PLL (29.48 MHz))
#pragma config WDTWIN = WDTWIN25        // Watchdog Window Select (Watchdog Window is 25% of WDT period)
#pragma config PWMPIN = ON //ON              // Motor Control PWM Module Pin Mode bit (PWM module pins controlled by PORT register at device Reset)
#pragma config PWMLOCK = OFF             // PWM Lock Enable (Certain PWM registers may only be written after key sequence)
#pragma config IESO = ON                // Internal External Switch Over Mode (Start-up device with FRC, then automatically switch to user-selected oscillator source when ready)

// CONFIG1
#pragma config WDTPOST = PS32768        // Watchdog Timer Postscaler (1:32,768)
#pragma config WDTPRE = PR128           // WDT Prescaler (Prescaler ratio of 1:128)
#pragma config PLLKEN = ON              // PLL Lock Enable (Clock switch to PLL source will wait until the PLL lock signal is valid.)
#pragma config WINDIS = OFF             // Watchdog Timer Window (Watchdog Timer in Non-Window mode)
#pragma config FWDTEN = OFF              // Watchdog Timer Enable (Watchdog timer always enabled)
#pragma config ICS = PGD1               // Comm Channel Select (Communicate on PGEC1/PGED1)
#pragma config HPOL = ON                // Motor Control PWM High Side Polarity bit (PWM module high side output pins have active-high output polarity)
#pragma config GWRP = OFF               // General Code Segment Write Protect (Writes to program memory are allowed)
#pragma config GCP = OFF                // General Segment Code Protection (General Segment Code protect is disabled)

// pragma config statements should precede project file includes.
// Use project enums instead of define for ON and OFF.

#include <xc.h>

#define FCY 14740000UL 
#include <libpic30.h>
#include <p33FJ16MC102.h>
#include    <math.h>

/*                          dsPIC33FJ16MC102
 *                   +------------:_:------------+
 *            RE3 -> :  1 VPP/MC              28 : <- AVdd
 *            RA0 <> :  2 AN0/PGED2           27 : <- AVss
 *            RA1 <> :  3 AN1/PGEC2    PWM1L1 26 : <> RB15
 *            RB0 <> :  4 AN2/PGED1    PWM1H1 25 : <> RB14
 *            RB1 <> :  5 AN3/PGEC1    PWM1L2 24 : <> RB13
 *            RB2 <> :  6 AN4/RP2      PWM1H2 23 : <> RB12
 *            RB3 <> :  7 AN5/RP3      PWM1L3 22 : <> RB11
 *            Vss -> :  8              PWM1H3 21 : <> RB10
 *            RA2 <> :  9 OSCI/CLKI           20 : <- Vcap
 *            RA3 <> : 10 OSCO/CLKO           19 : <- VSS
 *            RB4 <> : 11 RP4/PGED3 SDA1/DSI1 18 : <> RB9
 *            RA4 <> : 12 PGEC3     SCL1/SDO1 17 : <> RB8
 *            Vdd -> : 13           SCK1/INT0 16 : <> RB7
 *            RB5 <> : 14 RP5/FLTB1 RP6/FLTA1 15 : <> RB6
 *                   +---------------------------+
 *                              DIP-28
*/ 

#define NOMFREQ     60
#define FREQSTEP    1
#define MINFREQ     10
#define MAXFREQ     120

#define NOMVOLTS    240
#define VOLTSTEP    1
#define MINVOLTS    10
#define MAXVOLTS    300

#define LOOPTIME    50

#define MAXTABLE    361
#define PI          3.1416
//#define MAXPWM      1000
//#define DEBUG
//#define OFFSET      1000

/******************************************************************************/
/***** GLOBAL VARIABLES *******************************************************/
/******************************************************************************/

float                   phi = 0; // Phase
volatile float          increment;
volatile unsigned int   frequency = 60;
unsigned int            carrier = 8000;
volatile unsigned int   msec = 0;
volatile int            sine[MAXTABLE];
volatile unsigned int   Vadj = 200;

//char quad = 1; // Quadrant
//unsigned int    timebase; // = 240*MAXPWM/frequency; // 4000 -> 60 Hz

// Sine wave converted in a duty cycle in 15 points (6 degrees per point)
// for the 1st quadrant only: [0-90] (limits included) 
// Max PWM is 1000, with offset of 500
/*const unsigned short sin100_0_90[16] = { 
    500, 552, 604, 655, 708, 735, 750, 835, 871, 905, 933, 957, 975, 989, 997, 1000 
};
/*const unsigned short sin100_0_90[16] = { 
    250, 226, 302, 327, 354, 367, 375, 418, 436, 453, 467, 479, 488, 495, 498, 500
};*/

/* This version calculates the required duty cycle based on PWM carrier frequency
 * It requires some combination of single precision sine function or sinf()
 * as well as settings for fast math, and optimization level of 2
 */

int main(void) {
    
/******************************************************************************/
/***** INITIALIZATION: OSCILLATOR *********************************************/
/******************************************************************************/
    
    CLKDIVbits.FRCDIV = 0; // 29.48 MHz / 1
    CLKDIVbits.DOZE = 0b000;    // 0b000->CLK/1, 0b010=>CLK/4
    CLKDIVbits.ROI = 0;   // Interrupts have no effect
    CLKDIVbits.DOZEN = 0;   // 1->Enable, 0-> default 1/1
    
    /*
    ** Initiate clock switch
    ** 1. Wait for clock switch to occur
    ** 2. Wait for PLL to lock
    */
    
    // Primary Oscillator with PLL (XTPLL, HSPLL, ECPLL)
 //   __builtin_write_OSCCONH(0x03); // 0b011 -> NOSC<2:0> (OSCCON<10:8>)
    
    // Request oscillator switch to selection specified by the NOSC<2:0> bits
 //   __builtin_write_OSCCONL(OSCCON | 0x01); // 0b1 -> OSWEN<0> (OSCCON<0>)
    
    // Wait for Clock switch to occur
    // Current Oscillator Selection bits (read-only): COSC<2:0> (OSCCON<15:13>)


    OSCCONbits.COSC = 0b001;    //FOSC with PLL
//    OSCCONbits.NOSC = 0b001;    //FOSC with PLL
//    OSCCONbits.OSWEN = 1;
//    while(OSCCONbits.OSWEN);    // wait until switch complete
    // Wait for PLL to lock
    // PLL Lock Status bit (read-only): LOCK<0> (OSCCON<5>)
//    while (OSCCONbits.LOCK!=1);
    
/******************************************************************************/
/***** INITIALIZATION: PORTS **************************************************/
/******************************************************************************/
    
    // Set all ports to low
    LATA = 0x00;
    LATB = 0x00;
    
    // Set all ports to digital
    AD1PCFGL = 0xffff;  // digital

    // Set all ports to digital output
    TRISA = 0x00;
    TRISB = 0x00;
    
/* Initialize sine table */
    int i;
    for(i=0; i<MAXTABLE; i++)
        sine[i] = (int)1000*sin((float)i * PI / 180);
    
/******************************************************************************/
/***** INITIALIZATION: PWM ****************************************************/
/******************************************************************************/
    
    // ACRONYMS:
    // PWM Pulse-Width Modulation
    // ITB Independent Time Base
    // CAM Center-Aligned Mode
    // CMM Complementary mode
    
    // PWM time base control register
    PTCONbits.PTEN = 0; // Disable high-speed PWM module
    PTCONbits.PTCKPS = 0b00; // 00-> PWM clock/1, 01->PWM clock /4 
    PTCONbits.PTMOD = 0b00; //00-> Free Run, 10-> Up/Down, 11-> Up/Dn w/int

// External fault input pins 14 and 15 must be pulled high, then faults cleared
// FLTA1 pin must be pulled high externally in order to clear and disable the Fault
// Writing to P1FLTACON register requires unlock sequence
// Use builtin function to write 0x0000 to P1FLTACON register
__builtin_write_PWMSFR(&P1FLTACON, 0x0000, &PWM1KEY);
// FLTB1 pin must be pulled high externally in order to clear and disable the Fault
// Writing to P1FLTBCON register requires unlock sequence
// Use builtin function to write 0x0000 to P1FLTBCON register
__builtin_write_PWMSFR(&P1FLTBCON, 0x0000, &PWM1KEY);
// Enable all PWMs using PWM1CON1 register
// Writing to PWM1CON1 register requires unlock sequence
// Use builtin function to write 0x0077 to PWM1CON1 register
__builtin_write_PWMSFR(&PWM1CON1, 0x0077, &PWM1KEY);   

// Enable PWM outputs    
    PWM1CON1bits.PEN1H = 1;
    PWM1CON1bits.PEN1L = 1;
    PWM1CON1bits.PEN2H = 1;
    PWM1CON1bits.PEN2L = 1;
    PWM1CON1bits.PEN3H = 1;
    PWM1CON1bits.PEN3L = 1;
// Complememntary mode 0, independent 1
    PWM1CON1bits.PMOD1 = 0;
    PWM1CON1bits.PMOD2 = 0;
    PWM1CON1bits.PMOD3 = 0;
    
    // Configure PWM1CON2
    PWM1CON2bits.OSYNC = 0;
    PWM1CON2bits.SEVOPS = 0b0011;   // 1:4 Postscale
    PWM1CON2bits.UDIS = 0; // Updates enabled
    PWM1CON2bits.IUE = 0; // Updates to the active PDCx/SDCx registers are synchronized to the local PWM time base
    

// Enable high-speed PWM
    PTCONbits.PTEN = 1;
#ifdef DEBUG  
    P1DC1 = MAXPWM;
    __delay_ms(5000);
    P1DC1 = 9*(MAXPWM/10);
    __delay_ms(5000);
    P1DC1 = MAXPWM/2;
    __delay_ms(5000);
     P1DC1 = MAXPWM/16;
    __delay_ms(5000);
#endif   
     IEC3bits.PWM1IE = 1;    // Enable PWM interrupt
     
/* Timer1 Setup */
     T1CONbits.TCS = 0; // Internal clock TCy 14740000
     T1CONbits.TCKPS = 0b00; // 01=1/16 or 921250, 10=1/64 or 230312.5, 11=1/256 or 57578
     PR1 = 14740; // Approximately 1 kHz
     IEC0bits.T1IE = 1;
     T1CONbits.TON = 1;
     
     
/******************************************************************************/
/***** INFINITE LOOP **********************************************************/
/******************************************************************************/
    frequency = NOMFREQ; //MINFREQ;
    carrier = 8000; // 5000 is OK, 6000 -> 40 Hz/60
    phi = 0;
    char countdir = 1;  // 1 for count up, -1 for count down
    P1TPER = FCY/(1*carrier); // 2 for center aligned up/down counter, 1 for free run
    while(1) {
        increment = 1 *(float)(frequency * 360) / carrier; // 1 for center aligned or free
        msec = 0;
//        if( OSCCONbits.CF )
//            msec = STEPTIME/16;
        while( msec < LOOPTIME); // Delay in msec
        LATA ^= 0x0002;  // Toggle GRN LED - causes output jitter?
        if(frequency < MINFREQ)
            countdir = 1;
        else if(frequency > MAXFREQ)
            countdir = -1;
        frequency += countdir * FREQSTEP;
        Vadj = frequency * MAXVOLTS / NOMFREQ;
/*        if( Vadj < MAXVOLTS )
            Vadj += VOLTSTEP;
        else
            Vadj = MINVOLTS;
 */
    }
}

/******************************************************************************/
/***** INTERRUPT SERVICE ROUTINE: PWM2 ****************************************/
/******************************************************************************/

// Sinewave is generated in this function

void __attribute__((interrupt, no_auto_psv)) _DefaultInterrupt(void) {
    if(IFS3bits.PWM1IF) {
    //    long    per = P1TPER;
        LATAbits.LATA0 = 1;  // RED LED on
        phi+=increment;
        if (phi >=360)
            phi=0;    //-=360;
    //    int p = (int)phi;
        P1DC1 = Vadj * (P1TPER/2 + ((long)P1TPER * sine[(int)phi]) / 2000) / NOMVOLTS ; // Phase A
        P1DC2 = Vadj * (P1TPER/2 + ((long)P1TPER * sine[((int)phi+120)%360] )/2000) / NOMVOLTS ; // Phase B
        P1DC3 = Vadj * (P1TPER/2 + ((long)P1TPER * sine[((int)phi+240)%360] )/2000) / NOMVOLTS ; // Phase C
        LATAbits.LATA0 = 0;  // RED LED off 52.5 /126 uSec (36 before Vadj?)
        IFS3bits.PWM1IF = 0; }
    else if(IFS0bits.T1IF) {
        msec++;
        IFS0bits.T1IF=0;
    }
}